# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import time

from collections import OrderedDict

from driverng import Worklist
from driverng.astm import AstmMessageV2, HeaderSegment
from driverng.astm import HprimDate as HL7Date
from driverng.astm import HprimForeign as HL7Foreign
from driverng.astm import HprimMultiple
from driverng.astm import HprimString as HL7String
from driverng.astm import Segment
from driverng.astm.fields import HprimList

logger = logging.getLogger("driverng.drivers.cobas_pure")


class CobasPureSegmentMSH(HeaderSegment):
    sending_application = HL7String(3)
    sending_facility = HL7String(4)
    receiving_application = HL7String(5)
    receiving_facility = HL7String(6)
    message_date = HL7Date(7)
    message_type = HprimMultiple(
        9,
        message_code=HL7String(1, default="OML"),
        trigger_event=HL7String(2, default="O33"),
        message_structure=HL7String(3, default="OML_O33"),
    )
    control_id = HL7String(10)
    processing_id = HL7String(11, default="P")
    version = HL7String(12, default="2.5.1")
    accept_ack_type = HL7String(15, default="NE")
    application_ack_type = HL7String(16, default="AL")
    character_set = HL7String(18, default="UNICODE UTF-8")
    entity_id = HL7String((21, 1), default="LAB-28R")
    namespace_id = HL7String((21, 2), default="ROCHE")

    equipments = HL7Foreign("EQU", "header ")
    msas = HL7Foreign("MSA", "header")
    patients = HL7Foreign("PID", "header")
    queries = HL7Foreign("QPD", "header")
    specimens = HL7Foreign("SPM", "header")
    specimen_containers = HL7Foreign("SAC", "header")


class CobasPureSegmentEQU(Segment):
    equipment_id = HL7String((2, 1))
    namespace_id = HL7String((2, 2))
    event_date = HL7Date(3)
    equipment_status = HL7String(4)

    inventory = HL7Foreign("INV", "equipment")


class CobasPureSegmentPID(Segment):
    patient_id = HL7String(4)
    name_type_code = HL7String((6, 7), default="U")
    birthdate = HL7Date(8, fmt_date="%Y%m%d")
    sex = HL7String(9)

    specimens = HL7Foreign("SPM", "patient")


class CobasPureSegmentSPM(Segment):
    spm_sample_id = HprimMultiple(3, identifier=HprimList(1))
    sample_id = HL7String((3, 1))
    specimen_type_identifier = HL7String((5, 1))
    type_coding_system = HL7String((5, 3))
    specimen_role_identifier = HL7String((12, 1), default="P")
    role_coding_system = HL7String((12, 3), default="HL70369")
    comments = HprimList(15, basefield=HprimMultiple(0, content=HL7String(1)))
    collection_date = HL7Date(18)
    expiration_date = HL7Date(20)
    specimen_condition = HL7String((25, 1))
    condition_coding_system = HL7String((25, 3))
    container_type = HL7String((28, 1))
    container_coding_system = HL7String((28, 3))

    results = HL7Foreign("OBX", "specimen")
    specimen_containers = HL7Foreign("SAC", "specimen")


class CobasPureSegmentSAC(Segment):
    sample_id = HL7String((4, 1))
    type = HL7String((4, 2))
    rack = HL7String(11)
    position = HL7String(12)
    pre_dilution_code = HL7String((30, 2))
    pre_dilution_value = HL7String((30, 3))

    common_order = HL7Foreign("ORC", "specimen")
    orders = HL7Foreign("OBR", "specimen")


class CobasPureSegmentOBR(Segment):
    placer_order_number = HL7String(3)
    identifier = HL7String((5, 1))
    test = HL7String((5, 2))
    coding_system = HL7String((5, 3), default="99ROC")
    action_code = HL7String(12)

    common_order = HL7Foreign("ORC", "order")
    results = HL7Foreign("OBX", "order")
    sample_priority = HL7Foreign("TQ1", "order")
    test_infos = HL7Foreign("TCD", "order")


class CobasPureSegmentORC(Segment):
    order_control = HL7String(2)
    order_status = HL7String(6)
    transaction_date = HL7Date(10)

    orders = HL7Foreign("OBR", "order")
    sample_priority = HL7Foreign("TQ1", "order")

    def to_string(self, separators=None):
        self.rang_segment = self.order_control
        return super(CobasPureSegmentORC, self).to_string(separators=separators)


class CobasPureSegmentTQ1(Segment):
    priority = HL7String((10, 1))
    coding_system = HL7String((10, 3), default="HL70485")

    orders = HL7Foreign("OBR", "order")


class CobasPureSegmentOBX(Segment):
    value_type = HL7String(3)
    observation_identifier = HL7String((4, 1))
    observation_text = HL7String((4, 2))
    coding_system = HL7String((4, 3))
    observation_sub_id = HL7String(5)
    value = HL7String(6)
    units = HL7String((7, 1))
    units_coding_system = HL7String((7, 3))
    flag_identifier = HL7String((9, 1))
    flag_text = HL7String((9, 2))
    flag_coding_system = HL7String((9, 3))
    result_status = HL7String(12)
    observer = HL7String(17)
    measurement_unit = HL7String((19, 1))
    manufacturer = HL7String((19, 2))
    analysis_date = HL7Date(20)
    observation_type = HL7String(30)

    inventory = HL7Foreign("INV", "result")
    test_infos = HL7Foreign("TCD", "result")


class CobasPureSegmentTCD(Segment):
    universal_service_identifier = HprimMultiple(
        2, identifier=HL7String(1), coding_system=HL7String(3, default="99ROC")
    )
    num_1 = HL7String((3, 2))
    suffix = HL7String((3, 3))
    dilution_factor = HL7String((3, 4))

    inventory = HL7Foreign("INV", "order")

    def to_string(self, separators=None):
        self.rang_segment = self.universal_service_identifier
        return super(CobasPureSegmentTCD, self).to_string(separators=separators)


class CobasPureSegmentINV(Segment):
    substance_identifier = HL7String((2, 1))
    substance_status = HL7String((3, 1))
    coding_system = HL7String((3, 3))
    reagent_type = HL7String(4)
    sequence_number = HL7String(5)
    container_carrier_identifier = HL7String(6)
    position = HL7String(7)
    expiry = HL7Date(13)
    lot = HL7String(17)


class CobasPureSegmentQPD(Segment):
    message_identifier = HL7String((2, 1))
    message_text = HL7String((2, 2))
    coding_system = HL7String((2, 3))
    query_tag = HL7String(3)
    sample_identifier = HL7String(4)
    rack = HL7String(5)
    position = HL7String(6)
    sample_type = HL7String((11, 1))
    type_coding_system = HL7String((11, 3))
    sample_container_type = HL7String((12, 1))
    container_type_coding_system = HL7String((12, 3))
    priority = HL7String(13)
    query_kind = HL7String(14)

    queries_infos = HL7Foreign("RCP", "queries")


class CobasPureSegmentRCP(Segment):
    query_priority = HL7String(2)
    query_limited_request = HL7String(3)
    response_modality = HL7String((4, 1))
    coding_system = HL7String((4, 3))


class CobasPureSegmentMSA(Segment):
    ack_code = HL7String(2)
    control_id = HL7String(3)


class Driver(AstmMessageV2):
    SPEC = {"code": "cobas_pure", "label": "Cobas Pure"}
    ENCODING = "UTF-8"

    SEGMENT_CLASS_MAP = {
        "MSH": CobasPureSegmentMSH,
        "EQU": CobasPureSegmentEQU,
        "PID": CobasPureSegmentPID,
        "SPM": CobasPureSegmentSPM,
        "SAC": CobasPureSegmentSAC,
        "OBR": CobasPureSegmentOBR,
        "ORC": CobasPureSegmentORC,
        "TQ1": CobasPureSegmentTQ1,
        "OBX": CobasPureSegmentOBX,
        "TCD": CobasPureSegmentTCD,
        "INV": CobasPureSegmentINV,
        "QPD": CobasPureSegmentQPD,
        "RCP": CobasPureSegmentRCP,
        "MSA": CobasPureSegmentMSA,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    def create(self, **kwargs):
        return CobasPureSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        message = self.create(
            control_id=context.get("control_id", int(time.time() * 1000)),
            message_date=context["now"],
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "CA", worklist.ELSE: "NW"},
        ):
            seg_pid = message.patients.add(
                patient_id=patient.patient_id,
                birthdate=patient.birthdate if patient.birthdate else "",
                sex=patient.sex if patient.sex else "",
            )
            for sample, action, priority in patient.samples:
                specimen_type = "SERPLAS"
                coding_system = "99ROC"
                if sample.specimen == "Urine":
                    specimen_type = "UR"
                    coding_system = "HL70487"
                elif sample.specimen == "Whole blood":
                    specimen_type = "WB"
                    coding_system = "HL70487"
                seg_spm = seg_pid.specimens.add(
                    specimen_type_identifier=specimen_type,
                    type_coding_system=coding_system,
                    container_type="SC",
                    container_coding_system="99ROC",
                )
                seg_spm.spm_sample_id.identifier = [sample.sample_id, "BARCODE"]
                rack = ""
                position = ""
                if context["query"]:  # renvoi du container, du rack et de la position
                    for seg_qpd in context["query"].queries:
                        seg_spm.container_type = seg_qpd.sample_container_type
                        seg_spm.container_coding_system = (
                            seg_qpd.container_type_coding_system
                        )
                        rack = seg_qpd.rack
                        position = seg_qpd.position
                spm_comments = OrderedDict()
                if patient.last_name:
                    spm_comments["last_name"] = patient.last_name
                else:
                    spm_comments["last_name"] = ""
                if patient.first_name:
                    spm_comments["first_name"] = patient.first_name
                else:
                    spm_comments["first_name"] = ""
                if sample.patient_id:
                    spm_comments["ipp"] = sample.patient_id
                else:
                    spm_comments["ipp"] = ""
                if patient.maiden_name:
                    spm_comments["birth_name"] = patient.maiden_name
                else:
                    spm_comments["birth_name"] = spm_comments["last_name"]
                prescriber = [p for p in patient.physicians if p.type == "prescriber"]
                if prescriber:
                    spm_comments["prescriber"] = prescriber[0].last_name
                else:
                    spm_comments["prescriber"] = ""
                for key in spm_comments:
                    comment = seg_spm.comments.add()
                    comment.content = spm_comments.get(key)
                seg_sac = seg_spm.specimen_containers.add(
                    sample_id=sample.sample_id,
                    rack=rack,
                    position=position,
                    pre_dilution_code="1",
                    pre_dilution_value="1",
                )
                for order in sample.orders:
                    seg_orc = seg_sac.common_order.add(
                        order_control=action, transaction_date=sample.request_date
                    )
                    seg_tq1 = seg_orc.sample_priority.add(priority=priority)
                    seg_obr = seg_tq1.orders.add(
                        placer_order_number=sample.request_id, identifier=order.code
                    )
                    seg_tcd = seg_obr.test_infos.add()
                    seg_tcd.universal_service_identifier.identifier = order.code
                    seg_tcd.num_1 = "1"
                    seg_tcd.suffix = ":"
                    seg_tcd.dilution_factor = "1"
                    if order.dilution and ":" in order.dilution:
                        seg_tcd.dilution_factor = order.dilution.rsplit(":")[1]
        if message.patients:
            return message
        elif context["query"]:
            for seg_qpd in context["query"].queries:
                seg_spm = message.specimens.add(
                    specimen_type_identifier="SERPLAS"
                    if seg_qpd.sample_type == '""'
                    else seg_qpd.sample_type,
                    type_coding_system="99ROC"
                    if seg_qpd.sample_type == '""'
                    else seg_qpd.type_coding_system,
                    container_type=seg_qpd.sample_container_type,
                    container_coding_system="99ROC",
                )
                seg_spm.spm_sample_id.identifier = [
                    seg_qpd.sample_identifier,
                    "BARCODE",
                ]
                seg_sac = seg_spm.specimen_containers.add(
                    sample_id=seg_qpd.sample_identifier,
                    type="BARCODE",
                    rack=seg_qpd.rack,
                    position=seg_qpd.position,
                    pre_dilution_code="1",
                    pre_dilution_value="1",
                )
                seg_orc = seg_sac.common_order.add(
                    order_control="DC", transaction_date=context["now"]
                )
            return message

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qpd in message.queries:
                sample = worklist.get_or_create(
                    type="patient", sample_id=seg_qpd.sample_identifier
                )
                if seg_qpd.rack and seg_qpd.position:
                    sample.rack = "{}-{}".format(seg_qpd.rack, seg_qpd.position)
        else:
            worklist = Worklist(context="result")
            if message.equipments:
                return worklist
            if message.patients:
                patients = message.patients
                for patient in patients:
                    if patient.specimens:
                        specimens = patient.specimens
            elif message.specimens:
                specimens = message.specimens
            for seg_spm in specimens:
                if seg_spm.specimen_role_identifier == "Q":
                    sample = worklist.get_or_create(
                        sample_id=seg_spm.sample_id,
                        control_lot=seg_spm.sample_id,
                        type="control",
                    )
                else:
                    sample = worklist.get_or_create(
                        sample_id=seg_spm.sample_id, type="patient"
                    )
                for seg_sac in seg_spm.specimen_containers:
                    if seg_sac.rack and seg_sac.position:
                        sample.rack = "{}-{}".format(seg_sac.rack, seg_sac.position)
                    for seg_obr in seg_sac.orders:
                        # liste de codes analyses à na pas répliquer si ils sont plusieurs fois présents dans différentes trames OBX
                        codes = []
                        for seg_obx in seg_obr.results:
                            if (
                                seg_obx.value_type == "NM"
                                and seg_obx.observation_identifier not in codes
                            ):
                                nm_result = sample.results.add(
                                    code=seg_obx.observation_identifier,
                                    value=seg_obx.value,
                                    unit=seg_obx.units,
                                    operator=seg_obx.observer,
                                    date=seg_obx.analysis_date,
                                    instrument=seg_obx.measurement_unit,
                                )
                                codes.append(nm_result.code)
                                nm_code = seg_obx.observation_identifier
                                flags = []
                                if seg_obx.flag_identifier:
                                    flags.append(seg_obx.flag_identifier)
                                nm_result.flags = flags
                            if (
                                seg_obx.value_type == "CE"
                                and seg_obx.observation_identifier == nm_code
                                and seg_obx.value != ""
                            ):
                                code = "{}_I".format(seg_obx.observation_identifier)
                                if code not in codes:
                                    ce_result = sample.results.add(
                                        code="{}_I".format(
                                            seg_obx.observation_identifier
                                        ),
                                        value=seg_obx.value,
                                        operator=seg_obx.observer,
                                        date=seg_obx.analysis_date,
                                        instrument=seg_obx.measurement_unit,
                                    )
                                    codes.append(ce_result.code)
                            if seg_obx.test_infos:
                                for seg_tcd in seg_obx.test_infos:
                                    inventory = seg_tcd.inventory
                            else:
                                inventory = seg_obx.inventory
                            for seg_inv in inventory:
                                nm_result.lots.add(
                                    type="reagent",
                                    name="{}-{}".format(
                                        seg_inv.substance_identifier,
                                        seg_inv.reagent_type,
                                    ),
                                    lot=seg_inv.lot,
                                    expiration=seg_inv.expiry,
                                )

        return worklist
