# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import time

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.fanci")


class FanciSegmentMSH(HeaderSegment):
    sending_application = HprimString(3)
    sending_facility = HprimString(4, default="POC")
    receiving_facility = HprimString(6, default="FBT")
    date = HprimDate(7)
    message_type = HprimString((9, 1), default="ORM")
    message_event = HprimString((9, 2), default="O01")
    control_id = HprimString(10)
    processing_id = HprimString(11, default="P")
    version_id = HprimString(12, default="2.3")
    accept_ack_type = HprimString(15, default="AL")
    application_ack_type = HprimString(16)

    patients = HprimForeign("PID", "header")
    orders = HprimForeign("OBR", "header")


class FanciSegmentPID(Segment):
    patient_id = HprimString(4)
    patient_name = HprimString(6)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)

    visits = HprimForeign("PV1", "patient")
    orders = HprimForeign("OBR", "patient")
    common_order = HprimForeign("ORC", "patient")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(FanciSegmentPID, self).to_string(separators=separators)


class FanciSegmentPV1(Segment):
    referring_doctor = HprimString(8)
    visit_number = HprimString(19)

    orders = HprimForeign("OBR", "patient")
    common_order = HprimForeign("ORC", "patient")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(FanciSegmentPV1, self).to_string(separators=separators)


class FanciSegmentORC(Segment):
    order_status = HprimString(6, default="SC")
    date = HprimDate(10)

    def to_string(self, separators=None):
        self.rang_segment = "NW"
        return super(FanciSegmentORC, self).to_string(separators=separators)


class FanciSegmentOBR(Segment):
    universal_test_id = HprimString((5, 4))
    test_label = HprimString((5, 5))
    requested_date = HprimDate(6)
    result_status = HprimString(25, default="F")

    results = HprimForeign("OBX", "order")
    common_order = HprimForeign("ORC", "order")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(FanciSegmentOBR, self).to_string(separators=separators)


class FanciSegmentOBX(Segment):
    value_type = HprimString(2, default="NM")
    observation_identifier = HprimString(3)
    test_code = HprimString((4, 4))
    test_label = HprimString((4, 5))
    test_flag = HprimString((4, 6))
    observation_value = HprimString(6)
    units = HprimString(7)
    result_status = HprimString(11)
    date = HprimDate(15)


class FanciSegmentQRD(Segment):
    format_code = HprimString(2)
    priority = HprimString(3)
    quantity_limited_request = HprimString(7)
    who_subject_filter = HprimString(8)
    what_subject_filter = HprimString(9)


class Driver(AstmMessageV2):
    SPEC = {"code": "fanci", "label": "FANci"}
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    SEGMENT_CLASS_MAP = {
        "MSH": FanciSegmentMSH,
        "PID": FanciSegmentPID,
        "PV1": FanciSegmentPV1,
        "ORC": FanciSegmentORC,
        "OBR": FanciSegmentOBR,
        "OBX": FanciSegmentOBX,
        "QRD": FanciSegmentQRD,
    }

    def create(self, **kwargs):
        return FanciSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        msg = self.create(
            date=context["now"],
            control_id=context.get("control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        ):
            for sample, action in patient.samples:
                seg_pid = msg.patients.add(
                    patient_id=sample.sample_id,
                    patient_name="{} {}".format(patient.last_name, patient.first_name),
                    birthdate=patient.birthdate,
                    sex=patient.sex,
                )
                seg_pid.common_order.add(date=context["now"])
                for order in sample["orders"]:
                    seg_pid.orders.add(
                        universal_test_id=order.code, requested_date=sample.request_date
                    )
        if msg.patients:
            return msg

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for seg_pid in message.patients:
            for seg_obr in seg_pid.orders:
                sample = worklist.get_or_create(
                    sample_id=seg_pid.patient_id, type="patient"
                )
                commentaire_criteria = []
                list_code_results = ("12CO2", "13CO2", "DELTA", "STATUS")
                dict_result = {}
                for seg_obx in seg_obr.results:
                    if seg_obx.test_code == "CRITER":
                        code_test = "{}_{}".format(
                            seg_obr.universal_test_id, seg_obx.test_code
                        )
                        commentaire_criteria.append(seg_obx.observation_value)
                    elif seg_obx.test_code == "MTABLE":
                        if (
                            len(seg_obx.observation_value) != 36
                            and len(seg_obx.observation_value) != 45
                        ):
                            continue
                        dict_result[list_code_results[0]] = (
                            seg_obx.observation_value[6:16].replace(",", ".", 1).strip()
                        )
                        dict_result[list_code_results[1]] = (
                            seg_obx.observation_value[16:26]
                            .replace(",", ".", 1)
                            .strip()
                        )
                        dict_result[list_code_results[2]] = (
                            seg_obx.observation_value[26:36]
                            .replace(",", ".", 1)
                            .strip()
                        )
                        if len(seg_obx.observation_value) == 45:
                            dict_result[list_code_results[3]] = (
                                seg_obx.observation_value[36:45]
                                .replace(",", ".", 1)
                                .strip()
                            )
                        for code, resu in sorted(dict_result.items()):
                            sample.results.add(
                                code="{}_{}_T{}_{}".format(
                                    seg_obr.universal_test_id,
                                    seg_obx.test_code,
                                    seg_obx.observation_value[:6]
                                    .replace(",", ".", 1)
                                    .strip(),
                                    code,
                                ),
                                value=resu,
                            )
                    else:
                        test = "{}_{}".format(
                            seg_obr.universal_test_id, seg_obx.test_code
                        )
                        sample.results.add(
                            code=test,
                            value=seg_obx.observation_value.replace(",", ".", 1),
                            date=seg_obx.date,
                        )
                if commentaire_criteria:
                    sample.results.add(
                        code=code_test, value="\r".join(commentaire_criteria)
                    )
        return worklist
