# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import math

from functools import cmp_to_key

from driverng.fixed import Date, FixedLengthFormat, Multiple, String
from driverng import JSONGraphContainer, MessageV2, Worklist

DOWN, UP = range(2)
logger = logging.getLogger("driverng.tosoh_rp")


class QueryInformation(FixedLengthFormat):
    FMT_DESCRIPTION = ["G,", String("sample_id", 20)]


class OrderInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "A",
        ",",
        String("sample_id", 20),
        ",",
        String("sample_type", 2, default="00"),
        String("container_type", 2, default="00"),
        String("reserved1", 2, default="  "),
        ",",
        String("a1c_item", 2, default="00"),
        String("reserved2", 6, default="      "),
        ",",
        String("reserved4", 8, default="        "),
    ]


class SampleInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "01",
        String("analyzer_name", 12),
        String("measurement_mode", 1, default="0"),
        String("measurement_unit", 1, default="1"),
        String("specimen_type", 1, default="0"),
        String("decimal_place_peak_area", 1),
        String("decimal_place_retention_time", 1),
        String("reserved", 4),
        String("specimen_id", 20),
        String("rack_no", 4),
        String("position_no", 2),
        String("operator", 10),
        Date("measurement_date", 12, fmt_date="%Y%m%d%H%M"),
    ]


class SampleNumber(FixedLengthFormat):
    FMT_DESCRIPTION = ["02", String("sample_number", 16)]


class MeasurementValue(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "03",
        String("component1", 6, right_justified=True),
        String("component2", 6, right_justified=True),
        String("component3", 6, right_justified=True),
        String("component4", 6, right_justified=True),
        String("component5", 6, right_justified=True),
        String("component6", 6, right_justified=True),
        String("component7", 6, right_justified=True),
        String("component8", 6, right_justified=True),
        String("component9", 6, right_justified=True),
        String("component10", 6, right_justified=True),
        String("component11", 6, right_justified=True),
        String("component12", 6, right_justified=True),
        String("total_area", 8, right_justified=True),
        String("reserved", 6, right_justified=True),
        String("hb1ac_jds", 6, right_justified=True),
        String("hb1ac_ngsp", 6, right_justified=True),
        String("hb1ac_monos", 6, right_justified=True),
        String("reserved1", 6, right_justified=True),
        String("reserved2", 6, right_justified=True),
        String("reserved3", 6, right_justified=True),
        String("reserved4", 6, right_justified=True),
        String("hb1ac_ifcc", 6, right_justified=True),
        String("calibrator_type", 2),
        String("theoretical_plate_number", 5, right_justified=True),
        String("print_configuration", 2),
    ]


class FlagInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "04",
        Multiple("flags", [String("flag_judgement", 2), String("flag_line_number", 2)]),
    ]


class DataCollectionCondition(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "05",
        String("collection_start_time", 4, right_justified=True),
        String("collection_end_time", 4, right_justified=True),
        String("collection_pitch", 4, right_justified=True),
        String("number_of_peaks", 2, right_justified=True),
        String("number_of_chromatographic_data", 4, right_justified=True),
    ]


class PeakData(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "06",
        String("peak_number", 2, right_justified=True),
        String("peak_name", 5),
        String("peak_shape", 1),
        String("base_start_point", 4, right_justified=True),
        String("base_end_point", 4, right_justified=True),
        String("peak_start_point", 4, right_justified=True),
        String("peak_top", 4, right_justified=True),
        String("peak_end_point", 4, right_justified=True),
        String("peak_area", 8, right_justified=True),
        String("peak_area_percent", 6, right_justified=True),
        String("retention_time", 6, right_justified=True),
    ]


class PeakDataEnd(FixedLengthFormat):
    FMT_DESCRIPTION = ["07"]


class RawData(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "08",
        String("block_number", 3, right_justified=True),
        Multiple("block_signals", [String("value", 9, right_justified=True)]),
    ]


class CalibrationInformation(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "09",
        String("calibration_factor_a", 8, right_justified=True),
        String("calibration_factor_b", 8, right_justified=True),
        String("reserved1", 8, right_justified=True),
        String("reserved2", 8, right_justified=True),
        String("reserved3", 4),
        String("calibration_date", 8),
        String("calibration_lot", 8),
        String("calibration_operator", 10),
    ]


class AnalyzerStatus(FixedLengthFormat):
    FMT_DESCRIPTION = [
        "10",
        String("column_count", 5, right_justified=True),
        String("column_lot", 8),
        String("column_expiration_date", 8),
        String("column_load_date", 8),
        String("eluent1_lot", 8),
        String("eluent1_expiration_date", 8),
        String("eluent1_load_date", 8),
        String("eluent2_lot", 8),
        String("eluent2_expiration_date", 8),
        String("eluent2_load_date", 8),
        String("eluent3_lot", 8),
        String("eluent3_expiration_date", 8),
        String("eluent3_load_date", 8),
        String("hws_lot", 8),
        String("hws_expiration_date", 8),
        String("hws_load_date", 8),
    ]


class One(object):
    def __init__(self, name, klass):
        self.name = name
        self.klass = klass

    def get(self, **kwargs):
        return self.klass(**kwargs)


class ManyContainer(list):
    def __init__(self, klass):
        self.klass = klass
        list.__init__(self)

    def add(self, **kwargs):
        obj = self.klass(**kwargs)
        self.append(obj)
        return obj


class Many(One):
    def get(self):
        return ManyContainer(self.klass)


class _TosohRpMessage(object):
    FMT_MAP = {
        "G,": One("query", QueryInformation),
        "01": One("sample_information", SampleInformation),
        "02": One("sample_number", SampleNumber),
        "03": One("measurement_value", MeasurementValue),
        "04": One("flag_information", FlagInformation),
        "05": One("data_collection_condition", DataCollectionCondition),
        "06": Many("peak_data", PeakData),
        "07": One("peak_data_end", PeakDataEnd),
        "08": Many("raw_data", RawData),
        "09": One("calibration_information", CalibrationInformation),
        "10": One("analyzer_status", AnalyzerStatus),
    }

    def __init__(self, **kwargs):
        for attr in self.FMT_MAP.values():
            setattr(self, attr.name, attr.get())
        super(_TosohRpMessage, self).__init__()

    @staticmethod
    def parse(data):
        obj = _TosohRpMessage()
        for line in data.splitlines():
            if line:
                attr = obj.FMT_MAP.get(line[:2])
                value = getattr(obj, attr.name)
                if isinstance(value, ManyContainer):
                    o = value.add()
                    o.parse(line)
                else:
                    o = value.parse(line)

        return obj

    def to_string(self):
        lines = []
        for key in range(1, 10):
            attr = self.FMT_MAP[str(key)]
            value = getattr(self, attr.name)
            if isinstance(value, ManyContainer):
                lines.extend([v.to_string() for v in value])
            else:
                lines.append(value.to_string())
        return "\n".join(lines)


class Driver(MessageV2):
    SPEC = {"code": "tosoh_rp", "label": "Tosoh RP+"}
    COMPONENT_MAP = {
        "0": (
            "FP%",
            "A1a%",
            "A1b%",
            "F%",
            "LA1c%",
            "SA1c%",
            "A0%",
            "",
            "",
            "",
            "",
            "",
            "Total area",
            "",
            "HbA1c (JDS)%",
            "HbA1c (NGSP)%",
            "HbA1c (MonoS)",
            "",
            "",
            "",
            "",
            "HbA1c (IFCC)mmol/mol",
            "Calibrator type",
            "Theoretical plate number",
            "Print configuration",
        ),
        "1": (
            "FP%",
            "A1a%",
            "A1b%",
            "F%",
            "LA1c%",
            "SA1c%",
            "A0%",
            "H-VAR",
            "",
            "",
            "",
            "",
            "Total area",
            "",
            "HbA1c (JDS)%",
            "HbA1c (NGSP)%",
            "HbA1c (MonoS)",
            "",
            "",
            "",
            "",
            "HbA1c (IFCC)mmol/mol",
            "Calibrator type",
            "Theoretical plate number",
            "Print configuration",
        ),
        "2": (
            "F%",
            "A0%",
            "A2%",
            "E%",
            "D%",
            "S%",
            "C%",
            "",
            "",
            "",
            "",
            "",
            "Total area",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "Theoretical plate number",
            "Print configuration",
        ),
    }

    OPTIONS = [
        MessageV2.Option(
            "tosoh_rp_sample_id_mode",
            label="Gestion du code barre",
            choices=(
                ("normal", "Normal"),
                ("ignore_first_2", "Ignorer les deux premiers caractères"),
            ),
            default="normal",
        ),
        MessageV2.BooleanOption(
            "tosoh_rp_no_rack", label="Ne pas intégrer le rack du tube", default=False
        ),
    ]

    def __init__(self, **kwargs):
        self.tosoh_rp_sample_id_mode = kwargs.pop("tosoh_rp_sample_id_mode", "normal")
        self.tosoh_rp_no_rack = kwargs.pop("tosoh_rp_no_rack", False)
        super(Driver, self).__init__(**kwargs)

    def create(self):
        return OrderInformation()

    def to_string(self, message, crlf="\r\n"):
        return message.to_string()

    def encode_worklist(self, worklist, context):
        message = self.create()
        if len(worklist.samples) > 1:
            logger.warning("Multiple samples order are unsupported")
        if worklist.samples:
            sample = worklist.samples[0]
            message.sample_id = sample.sample_id
            message.a1c_item = "01"
            if sample.specimen in ("00", "01"):
                message.container_type = sample["specimen"]
            dilutions = [o.dilution for o in sample.orders if o.dilution]
            if dilutions:
                message.sample_type = "01"
        else:
            if context.get("query", None):
                message.sample_id = context["query"].query.sample_id
            else:
                return None

        return message

    def parse(self, data):
        obj = _TosohRpMessage.parse(data)
        return obj

    def get_epaulements(self, points):
        atan = [math.atan(points[i] - points[i - 1]) for i in range(1, len(points))]
        epaulements = []
        for i in range(1, len(atan)):
            diff = atan[i] - atan[i - 1]
            if diff > 0:
                direction = UP
            else:
                direction = DOWN
            if not epaulements or epaulements[-1]["direction"] != direction:
                epaulements.append({"direction": direction, "x": i, "y": atan[i]})
        return epaulements

    def get_graph_analysis(self, graph):
        sa1c, a0 = None, None
        for peak in graph["peaks"]:
            if peak["name"] == "SA1C":
                sa1c = peak
            if peak["name"] == "A0":
                a0 = peak

        results = []

        if sa1c:
            # Épaulement gauche
            ep_sa1c_left = 0
            epaulements = self.get_epaulements(
                graph["points"][sa1c["start"] : sa1c["top"]]
            )
            for i in range(len(epaulements)):
                if [e["direction"] for e in epaulements[i : i + 3]] == [DOWN, UP, DOWN]:
                    ep_sa1c_left = max(
                        ep_sa1c_left,
                        abs(
                            epaulements[i + 1]["y"]
                            - min(epaulements[i]["y"], epaulements[i + 2]["y"])
                        ),
                    )

            # Épaulement droit
            ep_sa1c_right = 0
            epaulements = self.get_epaulements(
                graph["points"][sa1c["top"] : sa1c["end"]]
            )
            for i in range(len(epaulements)):
                if [e["direction"] for e in epaulements[i : i + 3]] == [UP, DOWN, UP]:
                    ep_sa1c_right = max(
                        ep_sa1c_right,
                        abs(
                            epaulements[i + 1]["y"]
                            - max(epaulements[i]["y"], epaulements[i + 2]["y"])
                        ),
                    )

            results.extend(
                [
                    {
                        "code": "A1C_EPA1G",
                        "value": str(round(math.degrees(ep_sa1c_left), 2)),
                    },
                    {
                        "code": "A1C_EPA1D",
                        "value": str(round(math.degrees(ep_sa1c_right), 2)),
                    },
                ]
            )

            # Hauteur en pourcentage du début SA1C par rapport à la
            # hauteur min du début du pic précédent jusqu'à la fin du
            # SA1C.
            # Permet de voir si le début du SA1C est un creux bien marqué.
            peaks = sorted(
                graph["peaks"], key=cmp_to_key(lambda a, b: a["start"] - b["start"])
            )
            names = [p["name"] for p in peaks]
            before_sa1c_index = names.index("SA1C") - 1
            if before_sa1c_index >= 0:
                peak = peaks[before_sa1c_index]
                vmin = min(graph["points"])
                points = [v - vmin for v in graph["points"]]
                height = points[sa1c["start"]]
                height_p = (height - min(points[peak["start"] : sa1c["end"]])) / height
                results.append(
                    {
                        "code": "A1C_START_SA1C_HEIGHT_P",
                        "value": str(round(height_p * 100, 2)),
                    }
                )

        if sa1c and a0:
            # Aire en pourcentage par rapport à l'aire totale.
            # de la fin de SA1C jusqu'au dernier point avant le top A0 ou
            # la courbe a une tangente avec un angle inférieur à 45°.
            vmin = min(graph["points"])
            points = [v - vmin for v in graph["points"]]
            part = points[sa1c["end"] : a0["top"]]
            last = None
            for i in range(1, len(part)):
                if part[i] - part[i - 1] < 1:  # atan(1) = 45°
                    last = sa1c["end"] + i

            total = sum(points)
            ep_a0_area = sum(points[sa1c["end"] : last])

            results.extend(
                [
                    {
                        "code": "A1C_EPA0_P",
                        "value": str(round((ep_a0_area / total) * 100, 2)),
                    }
                ]
            )

        return results

    def get_measurement_values(self, message):
        mode = message.sample_information.measurement_mode
        values = []
        values.append(message.measurement_value.component1)
        values.append(message.measurement_value.component2)
        values.append(message.measurement_value.component3)
        values.append(message.measurement_value.component4)
        values.append(message.measurement_value.component5)
        values.append(message.measurement_value.component6)
        values.append(message.measurement_value.component7)
        values.append(message.measurement_value.component8)
        values.append(message.measurement_value.component9)
        values.append(message.measurement_value.component10)
        values.append(message.measurement_value.component11)
        values.append(message.measurement_value.component12)
        values.append(message.measurement_value.total_area)
        values.append(message.measurement_value.reserved)
        values.append(message.measurement_value.hb1ac_jds)
        values.append(message.measurement_value.hb1ac_ngsp)
        values.append(message.measurement_value.hb1ac_monos)
        values.append(message.measurement_value.reserved1)
        values.append(message.measurement_value.reserved2)
        values.append(message.measurement_value.reserved3)
        values.append(message.measurement_value.reserved4)
        values.append(message.measurement_value.hb1ac_ifcc)
        values.append(message.measurement_value.calibrator_type)
        values.append(message.measurement_value.theoretical_plate_number)
        values.append(message.measurement_value.print_configuration)
        results = []
        for index, value in enumerate(values):
            try:
                test = self.COMPONENT_MAP[mode][index]
            except IndexError:
                continue
            if not test:
                continue
            results.append((test, value))
        return results

    def get_graphe(self, message):
        graph = {"graph_type": "a1c", "points": [], "peaks": []}
        graph["parameters"] = {
            "start_time": int(message.data_collection_condition.collection_start_time),
            "end_time": int(message.data_collection_condition.collection_end_time),
            "pitch": int(message.data_collection_condition.collection_pitch),
            "n_peaks": int(message.data_collection_condition.number_of_peaks),
            "n_chromatographic": int(
                message.data_collection_condition.number_of_chromatographic_data
            ),
        }
        for raw in message.raw_data:
            if (
                raw.block_signals[-2].value == "0.000"
                and raw.block_signals[-1].value == "0.000"
            ):
                raw.block_signals[-2].value = raw.block_signals[-3].value
                raw.block_signals[-1].value = raw.block_signals[-3].value
            for point in raw.block_signals:
                value = float(point.value)
                graph["points"].append(value)
        for peak in message.peak_data:
            graph["peaks"].append(
                {
                    "name": peak.peak_name,
                    "shape": peak.peak_shape,
                    "base_start_point": int(peak.base_start_point),
                    "base_end_point": int(peak.base_end_point),
                    "start": int(peak.peak_start_point),
                    "top": int(peak.peak_top),
                    "end": int(peak.peak_end_point),
                    "area": float(peak.peak_area),
                    "area_percent": float(peak.peak_area_percent),
                }
            )
        results = self.get_graph_analysis(graph)
        results.append(
            {
                "code": "GRAPHE_A1C",
                "value": "GRAPHE",
                "graph": JSONGraphContainer(graph),
            }
        )
        return results

    def get_legacy_graphe(self, message, alarmes):
        results = []
        results.append(
            {
                "code": "Number of peaks",
                "value": message.data_collection_condition.number_of_peaks,
                "flags": alarmes,
            }
        )
        total_area = sum([float(peak.peak_area) for peak in message.peak_data])
        results.append(
            {"code": "Total area", "value": "%.2f" % (total_area,), "flags": alarmes}
        )
        total_area_percent = sum(
            [float(peak.peak_area_percent) for peak in message.peak_data]
        )
        results.append(
            {
                "code": "Total area percent",
                "value": "%.2f" % (total_area_percent,),
                "flags": alarmes,
            }
        )
        total_retention_time = sum(
            [float(peak.retention_time) for peak in message.peak_data]
        )
        results.append(
            {
                "code": "Total retention time",
                "value": "%.2f" % (total_retention_time,),
                "flags": alarmes,
            }
        )

        for peak in message.peak_data:
            results.append(
                {
                    "code": "area_%s" % (peak.peak_name,),
                    "value": "%.2f" % (float(peak.peak_area),),
                    "flags": alarmes,
                }
            )
            results.append(
                {
                    "code": "areap_%s" % (peak.peak_name,),
                    "value": "%.2f" % (float(peak.peak_area_percent),),
                    "flags": alarmes,
                }
            )

        return results

    def get_sample_id(self, message):
        if self.tosoh_rp_sample_id_mode == "ignore_first_2":
            sample_id = message.sample_information.specimen_id[2:]
        elif self.tosoh_rp_sample_id_mode == "normal":
            sample_id = message.sample_information.specimen_id
        else:
            raise NotImplementedError(
                "tosoh_rp_sample_id_mode %s" % (self.tosoh_rp_sample_id_mode,)
            )

        return sample_id

    def decode_message(self, message):

        if message.query.sample_id:
            worklist = Worklist(context="query")
            worklist.get_or_create(
                type="patient", sample_id=message.query.sample_id.strip()
            )
            return worklist

        worklist = Worklist(context="result")
        if not message.sample_information.measurement_date:
            logger.warning("Data Record 01 rejected without 'measurement_date'")
            return None

        sample_id = self.get_sample_id(message)
        if message.sample_information.specimen_type == "2":
            sample = worklist.get_or_create(
                type="control", sample_id=sample_id, control_lot=sample_id
            )
            date = message.sample_information.measurement_date
            for test, value in self.get_measurement_values(message):
                sample.results.add(code=test, value=value, date=date)
            for peak in message.peak_data:
                if peak.peak_name == "SA1C":
                    sample.results.add(
                        code="areap_{}".format(peak.peak_name),
                        value="{}".format(float(peak.peak_area_percent)),
                        date=date,
                    )
        else:
            if self.tosoh_rp_no_rack:
                rack = ""
            else:
                rack = "{}-{}".format(
                    message.sample_information.rack_no,
                    message.sample_information.position_no,
                )
            sample = worklist.get_or_create(
                type="patient",
                sample_id=sample_id,
                rack=rack,
                request_mode_hbdrive=message.sample_information.measurement_mode,
            )
            date = message.sample_information.measurement_date

            for test, value in self.get_measurement_values(message):
                sample.results.add(
                    code=test,
                    value=value,
                    date=date,
                    operator=message.sample_information.operator,
                )
            for g in self.get_legacy_graphe(message, alarmes=[]) + self.get_graphe(
                message
            ):
                sample.results.add(g)

            flag_judgements = []
            flag_line_numbers = []
            for flag in message.flag_information.flags:
                if flag.flag_judgement != "00":
                    flag_judgements.append(flag.flag_judgement)
                if flag.flag_judgement != "00":
                    flag_line_numbers.append(flag.flag_line_number)
            sample.results.add(code="Flag Judgements", value=";".join(flag_judgements))
            sample.results.add(
                code="Flag Line Numbers", value=";".join(flag_line_numbers)
            )

        for peak in message.peak_data:
            sample.results.add(
                code="Retention Time Peak {}".format(peak.peak_name),
                value=peak.retention_time,
            )

        sample.results.add(
            code="Analyzer Name", value=message.sample_information.analyzer_name
        )
        sample.results.add(
            code="Measurement mode", value=message.sample_information.measurement_mode
        )
        sample.results.add(
            code="Measurement unit", value=message.sample_information.measurement_unit
        )
        sample.results.add(
            code="Specimen type", value=message.sample_information.specimen_type
        )
        sample.results.add(
            code="Decimal place of peak area%",
            value=message.sample_information.decimal_place_peak_area,
        )
        sample.results.add(
            code="Decimal place of Retention time",
            value=message.sample_information.decimal_place_retention_time,
        )
        sample.results.add(code="Rack No.", value=message.sample_information.rack_no)
        sample.results.add(
            code="Position No.", value=message.sample_information.position_no
        )
        sample.results.add(
            code="Operator identification", value=message.sample_information.operator
        )
        sample.results.add(
            code="Mesurement Date / Time",
            value=message.sample_information.measurement_date.strftime("%Y%m%d%H%M"),
        )
        sample.results.add(
            code="Sample number", value=message.sample_number.sample_number
        )
        sample.results.add(
            code="Collection start time",
            value=message.data_collection_condition.collection_start_time,
        )
        sample.results.add(
            code="Collection end time",
            value=message.data_collection_condition.collection_end_time,
        )
        sample.results.add(
            code="Collection pitch",
            value=message.data_collection_condition.collection_pitch,
        )
        sample.results.add(
            code="Number of chromatographic data",
            value=message.data_collection_condition.number_of_chromatographic_data,
        )

        if message.sample_information.measurement_mode == "2":
            sample.results.add(
                code="Calibration factor A for F",
                value=message.calibration_information.calibration_factor_a,
            )
            sample.results.add(
                code="Calibration factor B for F",
                value=message.calibration_information.calibration_factor_b,
            )
            sample.results.add(
                code="Calibration factor A for A2",
                value=message.calibration_information.reserved1,
            )
            sample.results.add(
                code="Calibration factor B for A2",
                value=message.calibration_information.reserved2,
            )
        else:
            sample.results.add(
                code="Calibration factor A",
                value=message.calibration_information.calibration_factor_a,
            )
            sample.results.add(
                code="Calibration factor B",
                value=message.calibration_information.calibration_factor_b,
            )
        sample.results.add(
            code="Calibration date",
            value=message.calibration_information.calibration_date,
        )
        sample.results.add(
            code="Calibration lot",
            value=message.calibration_information.calibration_lot,
        )
        sample.results.add(
            code="Calibration operator",
            value=message.calibration_information.calibration_operator,
        )
        sample.results.add(
            code="Column count", value=message.analyzer_status.column_count
        )
        sample.results.add(code="Column lot", value=message.analyzer_status.column_lot)
        sample.results.add(
            code="Column expiration date",
            value=message.analyzer_status.column_expiration_date,
        )
        sample.results.add(
            code="Column load date", value=message.analyzer_status.column_load_date
        )
        sample.results.add(
            code="Eluent1 lot", value=message.analyzer_status.eluent1_lot
        )
        sample.results.add(
            code="Eluent1 expiration date",
            value=message.analyzer_status.eluent1_expiration_date,
        )
        sample.results.add(
            code="Eluent1 load date", value=message.analyzer_status.eluent1_load_date
        )
        sample.results.add(
            code="Eluent2 lot", value=message.analyzer_status.eluent2_lot
        )
        sample.results.add(
            code="Eluent2 expiration date",
            value=message.analyzer_status.eluent2_expiration_date,
        )
        sample.results.add(
            code="Eluent2 load date", value=message.analyzer_status.eluent2_load_date
        )
        sample.results.add(
            code="Eluent3 lot", value=message.analyzer_status.eluent3_lot
        )
        sample.results.add(
            code="Eluent3 expiration date",
            value=message.analyzer_status.eluent3_expiration_date,
        )
        sample.results.add(
            code="Eluent3 load date", value=message.analyzer_status.eluent3_load_date
        )
        sample.results.add(code="HWS lot", value=message.analyzer_status.hws_lot)
        sample.results.add(
            code="HWS expiration date",
            value=message.analyzer_status.hws_expiration_date,
        )
        sample.results.add(
            code="HWS load date", value=message.analyzer_status.hws_load_date
        )

        return worklist
